//
// © 2014 David Vines
//
// Creative Commons License
//
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
//

/* globals Game: true, ROT */

Game.Builder = function(width, height, depth) {
    this._width = width;
    this._height = height;
    this._depth = depth;
    this._tiles = new Array(depth);
    this._regions = new Array(depth);
    
    for(var z=0; z<depth; z++) {
        this._tiles[z] = this._generateLevel();
        this._regions[z] = new Array(width);
        for(var x=0; x<width; x++) {
            this._regions[z][x] = new Array(height);
        }
        this._setupRegions(z);
    }
    
    this._connectAllRegions();
};

Game.Builder.prototype.getTiles = function() {
    return this._tiles;
};

Game.Builder.prototype.getDepth = function() {
    return this._depth;
};

Game.Builder.prototype.getWidth = function() {
    return this._width;
};

Game.Builder.prototype.getHeight = function() {
    return this._height;
};

Game.Builder.prototype._generateLevel = function() {
    var map = new Array(this._width);
    for (var x = 0; x < this._width; x++) {
        map[x] = new Array(this._height);
    }

    var generator = new ROT.Map.Cellular(this._width,this._height);
    generator.randomize(0.5);
    var totalIterations = 3;
    for (var i=0; i<totalIterations; i++) {
        generator.create();
    }
    generator.create(function(x,y,v) {
        if (v === 1) {
            map[x][y] = Game.Tile.floorTile;
        } else {
            map[x][y] = Game.Tile.wallTile;
        }
    });

    return map;
};

Game.Builder.prototype._canFillRegion = function(x, y, z) {
    if (x<0 || y<0 || z<0 || x>=this._width || y>=this._height || z>=this._depth) {
        return false;
    }
    if (this._regions[z][x][y]) {
        return false;
    }
    return this._tiles[z][x][y].isWalkable();
};

Game.Builder.prototype._fillRegion = function(region, x, y, z) {
    var tilesFilled = [];
    var tiles = [{x:x, y:y}];
    while(tiles.length !== 0) {
        var tile = tiles.pop();
        if (!this._regions[z][tile.x][tile.y]) {
            this._regions[z][tile.x][tile.y] = region;
            tilesFilled.push(tile);
            var neighbours = Game.getNeighbourPositions(tile.x,tile.y);
            while(neighbours.length !== 0) {
                tile = neighbours.pop();
                if (this._canFillRegion(tile.x, tile.y, z)) {
                    tiles.push(tile);
                }
            }
        }
    }
    return tilesFilled;
};

Game.Builder.prototype._removeRegion = function(tiles,z) {
    while(tiles.length !== 0) {
        var tile = tiles.pop();
        this._regions[z][tile.x][tile.y] = undefined;
        this._tiles[z][tile.x][tile.y] = Game.Tile.wallTile;
    }
};

Game.Builder.prototype._setupRegions = function(z) {
    var region = 1;
    for(var x=0; x<this._width; x++) {
        for(var y=0; y<this._height; y++) {
            if (this._canFillRegion(x, y, z)) {
                var tilesFilled = this._fillRegion(region, x, y, z);
                if (tilesFilled.length < 21) {
                    this._removeRegion(tilesFilled, z);
                } else {
                    region++;
                }
            }
        }
    }
};

Game.Builder.prototype._findRegionOverlaps = function(z, r1, r2) {
    var matches = [];
    for (var x=0; x<this._width; x++) {
        for (var y=0; y<this._height; y++) {
            if (  this._tiles[z][x][y] === Game.Tile.floorTile   &&
                  this._tiles[z+1][x][y] === Game.Tile.floorTile &&
                  this._regions[z][x][y] === r1                  &&
                  this._regions[z+1][x][y] === r2
               ) {
                matches.push({x:x, y: y});
            }
        }
    }
    return matches.randomize();
};

Game.Builder.prototype._connectRegions = function(z, r1, r2) {
    var overlap = this._findRegionOverlaps(z, r1, r2);
    if (overlap.length === 0) {
        return false;
    }
    var point = overlap[0];
    this._tiles[z][point.x][point.y] = Game.Tile.stairsDownTile;
    this._tiles[z+1][point.x][point.y] = Game.Tile.stairsUpTile;
    return true;
};

Game.Builder.prototype._connectAllRegions = function() {
    for(var z=0; z<this._depth-1; z++) {
        var attemptedConnection = [];
        for (var x=0; x<this._width; x++) {
            for(var y=0; y<this._height; y++) {
                var key = this._regions[z][x][y]+','+this._regions[z+1][x][y];
                if (  this._tiles[z][x][y] === Game.Tile.floorTile  &&
                      this._tiles[z+1][x][y] === Game.Tile.floorTile &&
                      !attemptedConnection[key]
                   ) {
                    this._connectRegions(z, this._regions[z][x][y], this._regions[z+1][x][y]);
                    attemptedConnection[key] = true;
                }
            }
        }
    }
};