//
// © 2014 David Vines
//
// Creative Commons License
//
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
//

/* globals Game: true, ROT */

Game.EntityMixins = {};

Game.EntityMixins.Moveable = {
    name: 'Moveable',
    init: function() {
        this._visitedTorian = false;
    },
    tryMove: function(x, y, z) {
        var tile = this._map.getTile(x,y,this.getZ());
        var target = this._map.getEntityAt(x,y,this.getZ());
        var farm = Game.getFarm();
        var cave = Game.getCave();
        var temple = Game.getTemple();
        if (z < this.getZ()) {
            if ((tile === Game.Tile.caveExitTile) && (this.hasMixin('PlayerActor'))) {
                Game.getCave().removeEntity(this);
                this.setX(farm.getWidth()-1);
                this.setY(farm.getHeight()/2);
                this.setZ(0);
                farm.addEntity(this);
                this.switchMap(farm);
                this.clearMessages();
                return true;
            } else if ((tile === Game.Tile.templeExitTile) && (this.hasMixin('PlayerActor'))) {
                Game.getTemple().removeEntity(this);
                cave.putAtTemple(this);
                this.switchMap(cave);
                this.clearMessages();
                return true;
            } else if (tile !== Game.Tile.stairsUpTile) {
                this.sendSelfMessage("You can't go up here!");
                return false;
            } else {
                this.sendSelfMessage("You ascend to level %s.",[z+1]);
                this.setPosition(x,y,z);
                return true;
            }
        } else if (z > this.getZ()) {
            if ((tile === Game.Tile.caveEntranceTile) && (this.hasMixin('PlayerActor'))) {
                Game.getFarm().removeEntity(this);
                if (!cave) {
                    cave = new Game.Map.Cave(new Game.Builder(100, 100, 4).getTiles(),this);
                    Game.setCave(cave);
                }
                cave.putAtEntrance(this);
                this.switchMap(cave);
                this.clearMessages();
                return true;
            } else if ((tile === Game.Tile.templeEntranceTile) && (this.hasMixin('PlayerActor'))) {
                Game.getCave().removeEntity(this);
                if (!temple) {
                    temple = new Game.Map.Temple();
                    Game.setTemple(temple);
                }
                temple.putAtEntrance(this);
                this.switchMap(temple);
                this.clearMessages();
                return true;
            } else if (tile !== Game.Tile.stairsDownTile) {
                this.sendSelfMessage("You can't go down here!");
                return false;
            } else {
                this.sendSelfMessage("You descend to level %s.",[z+1]);
                this.setPosition(x,y,z);
                return true;
            }
        } else if (  target &&
                     (this.hasMixin('PlayerActor') || target.hasMixin('PlayerActor'))
                  ) {
            if (this.hasMixin('Fighter') && target.hasMixin('Fighter')) {
                this.attack(target);
                return false;
            } else {
                return false;
            }
        } else if (target) {
            return false;
        }
        if (tile.isWalkable()) {
            this.setPosition(x,y,z);
            if (tile.getDescription()) {
                this.sendSelfMessage(tile.getDescription());
            }
            if (!this._visitedTorian && tile === Game.Tile.torianTile) {
                this.sendSelfMessage('This is the home of a kindly Torian priest. "I think I saw your sibling enter that cave to the east. You\'d better search there. Hmm it looks quite dangerous though, so you\'d better have these just in case." The Torian priest gives you some muesli, honey and mead.');
                this.addItem(new Game.Item(Game.TorianMuseliTemplate));
                this.addItem(new Game.Item(Game.TorianHoneyTemplate));
                this.addItem(new Game.Item(Game.TorianMeadTemplate));
                this._visitedTorian = true;
            }
            if (tile === Game.Tile.houseTile && this.hasMixin('PlayerActor') && this.hasMixin("InventoryHolder")) {
                for(var i=0; i<this.getItems().length; i++) {
                    if (this.getItem(i) === Game.SiblingItem) {
                        this.win();
                    }
                }
            }
            return true;
        } else if (tile.isDiggable() && this.hasMixin('PlayerActor')) {
            this._map.dig(x,y,z);
            return true;
        }
        return false;
    }
};

Game.EntityMixins.MessageReader = {
    name: 'MessageReader',
    init: function() {
        this._messages = [];
    },
    receiveMessage: function(message) {
        this._messages.push(message);
    },
    getMessages: function() {
        return this._messages;
    },
    clearMessages: function() {
        this._messages = [];
    }
};

Game.EntityMixins.Sight = {
    name: 'Sight',
    groupName: 'Sight',
    /* jshint sub:true */
    init: function(template) {
        this._sightRadius = template['sightRadius'] || 5;
        this._normalSightRadius = this._sightRadius;
    },
    getSightRadius: function() {
        if (this.hasMixin(Game.EntityMixins.Equipper)) {
            return this._sightRadius + this.getSwordHand().getAdditionalLight() + this.getShieldHand().getAdditionalLight();
        }
        return this._sightRadius;
    },
    canSee: function(entity) {
        if (!entity || this._map !== entity.getMap() || this._z !== entity.getZ()) {
            return false;
        }
        var otherx = entity.getX();
        var othery = entity.getY();
        var gapx = otherx - this._x;
        var gapy = othery - this._y;
        if (gapx*gapx+gapy*gapy > this.getSightRadius()*this.getSightRadius()) {
            return false;
        }
        var found = false;
        this.getMap().getFov(this.getZ()).compute(this.getX(),this.getY(),this.getSightRadius(),function(x,y,radius,visibility) {
            if (x === otherx && y === othery) {
                found = true;
            }
        });
        return found;
    },
    setSightRadius: function(radius) {
        this._sightRadius = radius;
    },
    getNormalSightRadius: function() {
        return this._normalSightRadius;
    }
};

Game.EntityMixins._BaseFighter = {
    attack: function(target) {
        if (this.hasMixin('Fighter') && target.hasMixin('Fighter')) {
            var recorder = null;
            var other = null;
            if (this.hasMixin('FightHandler')) {
                recorder = this;
                other = target;
            }
            if ( target.hasMixin('FightHandler') && ((recorder === null)||(target.getSerialNumber() < this.getSerialNumber())) ) {
                recorder = target;
                other = this;
            }
            
            if (recorder === null) {
               throw new Error("Neither "+this.getName()+" nor "+target.getName+" can record the need for this fight.");
            }
            
            recorder.addOpponent(other);
        } else {
            throw new Error("Both this "+this.getName()+" and target "+target.getName()+" need to be Fighter to have a fight");
        }
        target.raiseEvent("onAttacked");
        return false;
    }
};

Game.EntityMixins.PassiveFighter = {
    name: 'PassiveFighter',
    interfaceName: 'Fighter',
    _morale: 1,
    _modifier: 0,
    init: function() {
        this._hp = 1;
    },
    takeDamage: function(attacker, damage, locationMod) {
        if (damage === 0) {
            this.sendSelfMessage("You take no damage from the %s's attack.",[attacker.getName()]);
            this.sendMessage(attacker, "You fail to inflicit any damage on the %s.",[this.getName()]);
            return;
        } else {
            this.sendSelfMessage("You take %s damage from the %s's attack.",[damage,attacker.getName()]);
            this.sendMessage(attacker,"You cause %s damage to the %s",[damage,this.getName()]);
        }
        this._hp -= damage;
        if (this._hp <= 0) {
            this.sendMessage(attacker,"The %s has been killed.",[this.getName()]);
            this.kill("You have been killed.");
            return true;
        }
        return false;
    },
    takeMorale: function(attacker, damage) {
        // Simple's don't lose morale :)
        return false;
    },
    isDefensive: function() {
        return true;
    },
    attackPerformance: function() {
        return Game.Rules.performance(0,this._morale)+3+this._modifier;
    },
    getDamageScore: function() {
        return 0;
    },
    setModifier: function(mod) {
        this.modifier = mod;
    },
};
Game.EntityMixins.PassiveFighter.extend(Game.EntityMixins._BaseFighter);

Game.EntityMixins.HumanFighter = {
    name: 'HumanFighter',
    interfaceName: 'Fighter',
    _morale: null,
    _woundPoints: null,
    _totalWounds: 0,
    _modifier: 0,
    init: function() {
        this._morale = Game.Rules.startingMorale(this.sheet);
        this._woundPoints = Game.Rules.woundPoints(this.sheet);
    },
    _getArmour: function() {
        if (this.hasMixin(Game.EntityMixins.Equipper)) {
            return this.getArmour();
        } else {
            return Game.NullItem;
        }
    },
    _getDamage: function() {
        if (this.hasMixin(Game.EntityMixins.Equipper)) {
            return this.getSwordHand().getDamage();
        } else {
            return 0;
        }
    },
    _getSwordHandSkill: function() {
        if (this.hasMixin(Game.EntityMixins.Equipper)) {
            var skill = this.getSwordHand().getSkillName();
            return Game.Rules.skilllevel(skill,this.sheet,this.dps[skill]);
        } else {
            return 0;
        }
    },
    takeDamage: function(attacker, damage, locationMod) {
        if (damage === 0) {
            this.sendSelfMessage("You take no damage from the %s's attack.",[attacker.getName()]);
            this.sendMessage(attacker, "You fail to inflicit any damage on the %s.",[this.getName()]);
            return false;
        }
        var location = Game.Rules.meleeLocation(locationMod);
        damage -= this._getArmour()[location];
        if (damage > 0) {
            if (damage === 1) {
                this.sendSelfMessage("You take a wound point in the %s.",[location]);
                this.sendMessage(attacker,"You hit the %s for a wound point in the %s.",[this.getName(),location]);
            } else {
                this.sendSelfMessage("You take %s wound points in the %s.",[damage,location]);
                this.sendMessage(attacker,"You hit the %s for %s wound points in the %s.",[this.getName(),damage,location]);
            }
            this._woundPoints[location] -= damage;
            this._totalWounds += damage;
            if (this._woundPoints[location] <= 0) {
                this.sendSelfMessage("Your %s now has no wound points left.",[location]);
                this.sendMessage(attacker,"The %s's %s now has no wound points left.",[this.getName(),location]);
                var conroll = Game.Rules.performance(this.sheet.con,100); // Always uncapped
                this.sendSelfMessage("You make %s con survival roll.",[Game.Rules.aadj(conroll)]);
                this.sendMessage(attacker,"%s makes %s con survival roll.",[this.getName(),Game.Rules.aadj(conroll)]);
                if (conroll < this._totalWounds) {
                    this.sendMessage(attacker,"and hence failed the roll (rolled %s, but had a total of %s wound points) and died!",[conroll,this._totalWounds]);
                    this.kill("You failed the roll (rolled %s, but had a total of %s wound points) and hence have died.",[conroll,this._totalWounds]);
                    return true;
                }
            }
        } else {
            this.sendSelfMessage("Your %s armour absorbs all the damage from the %s.",[location,attacker.getName()]);
            this.sendMessage(attacker,"%s's %s armour absorbs all the damage from your attack.",[this.getName(), location]);
        }
        return false;
    },
    takeMorale: function(attacker, damage) {
        this._morale -= damage;
        this.sendSelfMessage("You lose %s morale.",[damage]);
        this.sendMessage(attacker,"%s lost %s morale.",[this.getName(), damage]);
        if (this._morale < 1) {
            this.sendMessage(attacker,"%s has lost consciousness and you put it out of its misery.",[this.getName()]);
            this.kill("You lose consciousness and die!");
            return true;
        }
        return false;
    },
    isDefensive: function() {
        return false;
    },
    attackPerformance: function() {
        return Game.Rules.performance(this._getSwordHandSkill(),this._morale)+this._modifier-this._getArmour().getEncumbrance();
    },
    getDamageScore: function() {
        return Game.Rules.damageScore(this._getDamage());
    },
    setModifier: function(mod) {
        this._modifier = mod;
    },
    getWoundPoints: function() {
        return this._woundPoints;
    },
    getTotalWounds: function() {
        return this._totalWounds;
    },
    getMorale: function() {
        return this._morale;
    },
    isDead: function() {
        return this._dead;
    },
    listeners: {
        onHealMorale: function(item) {
            if (this._morale < Game.Rules.startingMorale(this.sheet)) {
                this._morale = Game.Rules.startingMorale(this.sheet);
                this.sendSelfMessage("You regain your morale.");
            } else {
                this.sendSelfMessage("Using the "+item.getName()+" has no effect.");
            }
        },
        onHeal: function(item) {
            if (this._totalWounds > 0) {
                if (item.getHeals() === "all") {
                    this._woundPoints = Game.Rules.woundPoints(this.sheet);
                    this._totalWounds = 0;
                    this.sendSelfMessage("All your wounds are healed.");
                } else {
                    var start = Game.Rules.woundPoints(this.sheet);
                    var locations = ["leftLeg", "rightLeg", "abdomen", "chest", "swordArm" , "shieldArm", "head" ];
                    var worstLoc = "head";
                    var worstDam = start[worstLoc] - this._woundPoints[worstLoc];
                    for(var i=1; i<locations.length; i++) {
                        if ((start[locations[i]]-this._woundPoints[locations[i]]) > worstDam) {
                            worstLoc = locations[i];
                            worstDam = start[locations[i]]-this._woundPoints[locations[i]];
                        }
                    }
                    this._woundPoints[worstLoc] = start[worstLoc];
                    this._totalWounds -= worstDam;
                    this.sendSelfMessage("Your "+worstLoc+" is healed.");
                }
            } else {
                this.sendSelfMessage("Using the "+item.getName()+" has no effect.");
            }
        }
    }
};
Game.EntityMixins.HumanFighter.extend(Game.EntityMixins._BaseFighter);

Game.EntityMixins.SimpleFighter = {
    name: 'SimpleFighter',
    interfaceName: 'Fighter',
    _morale: null,
    _woundPoints: null,
    _skill: null,
    _damage: null,
    _modifier: 0,
    /* jshint sub:true */
    init: function(template) {
        this._morale = template['morale'] || 150;
        this._woundPoints = template['woundPoints'] || 10;
        this._skill = template['skill'] || 10;
        this._damage = template['damage'] || 5;
        
        // Adjust for rule set
        if (Game.Rules.formUnderlyingSelected() === "2014a") {
            this._skill = this._skill/5;
        }
    },
    takeDamage: function(attacker, damage, locationMod) {
        if (damage === 0) {
            this.sendSelfMessage("You take no damage from the %s's attack.",[attacker.getName()]);
            this.sendMessage(attacker, "You fail to inflicit any damage on the %s.",[this.getName()]);
            return false;
        }
        if (damage > 0) {
            if (damage === 1) {
                this.sendSelfMessage("You take a wound point.");
                this.sendMessage(attacker,"You hit the %s for a wound point.",[this.getName()]);
            } else {
                this.sendSelfMessage("You take %s wound points.",[damage]);
                this.sendMessage(attacker,"You hit the %s for %s wound points.",[this.getName(),damage]);
            }
            this._woundPoints -= damage;
            if (this._woundPoints <= 0) {
                this.sendMessage(attacker,"The %s now has no wound points left and has died.",[this.getName()]);
                this.kill("You now have no wound points left and have died.");
                return true;
            }
        }
        return false;
    },
    takeMorale: function(attacker, damage) {
        this._morale -= damage;
        this.sendSelfMessage("You lose %s morale.",[damage]);
        this.sendMessage(attacker,"The %s lost %s morale.",[this.getName(), damage]);
        if (this._morale < 1) {
            this.sendMessage(attacker,"%s has lost consciousness and you put it out of its misery.",[this.getName()]);
            this.kill("You lose consciousness and die!");
            return true;
        }
        return false;
    },
    isDefensive: function() {
        return false;
    },
    attackPerformance: function(cap) {
        return Game.Rules.performance(this._skill,this._morale)+this._modifier;
    },
    getDamageScore: function() {
        return Game.Rules.damageScore(this._damage);
    },
    setModifier: function(mod) {
        this._modifier = mod;
    }
};
Game.EntityMixins.SimpleFighter.extend(Game.EntityMixins._BaseFighter);

Game.EntityMixins.InventoryHolder = {
    name: 'InventoryHolder',
    init: function(template) {
        var inventorySlots = template['inventorySlots'] || 10;
        this._items = new Array(inventorySlots);
    },
    getItems: function() {
        return this._items;
    },
    getItem: function(i) {
        return this._items[i];
    },
    addItem: function(item) {
        for(var i=0; i<this._items.length; i++) {
            if (!this._items[i]) {
                this._items[i] = item;
                return true;
            }
        }
        return false;
    },
    removeItem: function(i) {
        if (this._items[i] && this.hasMixin(Game.EntityMixins.Equipper)) {
            this.unequip(this._items[i]);
        } 
        this._items[i] = null;
    },
    canAddItem: function() {
        for(var i=0; i<this._items.length; i++) {
            if (!this._items[i]) {
                return true;
            }
        }
        return false;
    },
    pickupItems: function(indicies) {
        var mapItems = this._map.getItemsAt(this.getX(),this.getY(),this.getZ());
        var added = 0;
        for(var i=0; i<indicies.length; i++) {
            if (this.addItem(mapItems[indicies[i] - added])) {
                mapItems.splice(indicies[i] - added, 1);
                added++;
            } else {
                break;
            }
        }
        
        this._map.setItemsAt(this.getX(), this.getY(), this.getZ(), mapItems);
        
        return added === indicies.length;
    },
    dropItem: function(i) {
        if (this._items[i]) {
            this._map.addItem(this.getX(), this.getY(), this.getZ(), this._items[i]);
            this.removeItem(i);
        }
    },    
    use: function(i) {
        if (this._items[i]) {
            var item = this._items[i];
            this.removeItem(i);
            this.raiseEvent("onUse",item);
        }
    }
};

Game.EntityMixins.Learner = {
    name: 'Learner',
    dps: {Knife:50, Sword:0, Club:75, Brawl:100},
    listeners: {
        onKill: function(points) {
            var skill = this.getSwordHand().getSkillName();
            this.dps[skill] += points;
            this.sendSelfMessage("You gain %s development points in %s.",[points,skill]);
        }
    }
};

Game.EntityMixins.PlayerActor = {
    name: 'PlayerActor',
    interfaceName: 'Actor',
    sheet: null,
    turns: 0,
    won: false,

    init: function() {
        this.sheet  = Game.Rules.newCharacter();
    },

    initialAct: function() {
        this.sendSelfMessage('"Your younger sibling has gone missing - can you go and look for them?"');
        this._map.describeItemsWith(this);
    },
    
    act: function() {
        this.turns++;
        this._map.describeItemsWith(this);
        this.resolveFights();
        if (!this.isAlive()) {
            Game.Screen.playScreen.setGameEnded();
            this.sendSelfMessage("You have died... Press [Enter] to continue!");
        }
        if (this.won) {
            Game.Screen.playScreen.setGameWon();
            this.sendSelfMessage("Congratulations! You have found and returned with your younger sibling.");
            this.sendSelfMessage("Press [Enter] to continue!");
        }
        Game.refresh();
        this.clearMessages();
        this.getMap().getEngine().lock();
    },
    
    win: function() {
        this.won = true;
    }
};

Game.EntityMixins.PossessedHuman = {
        name: 'PossessedHuman',
        sheet: null,
        dps: {Knife:1500, Sword:1000, Club:1750, Brawl:2000},

        init: function() {
            this.sheet  = Game.Rules.newCharacter();
            var sword = new Game.Item(Game.SwordTemplate);
            this.addItem(sword);
            this.wieldSword(sword);
        },
    };

Game.EntityMixins.FungusActor = {
    name: 'FungusActor',
    interfaceName: 'Actor',
    init: function() {
        this._growthsRemaining = 5;
    },
    act: function() {
        if (this._growthsRemaining > 0) {
            if (Math.random() < 0.02) {
                var xOffset = [-1,0,1].random();
                var yOffset = [-1,0,1].random();
                if (xOffset !== 0 || yOffset !== 0) {
                    if (this.getMap().isEmptyFloor(this.getX()+xOffset,this.getY()+yOffset,this.getZ())) {
                        var entity = new Game.Entity(Game.FungusTemplate);
                        entity.setPosition(this.getX()+xOffset,this.getY()+yOffset,this.getZ());
                        this.getMap().addEntity(entity);
                        this._growthsRemaining--;
                        entity.sendMessageNearby('The fungus is spreading');
                    }
                }
            }
        }
    }
};

Game.EntityMixins.FightHandler = {
    name: 'FightHandler',
    init: function() {
        this._opponents = [];
    },
    addOpponent: function(opponent) {
        var alreadyPresent = false;
        for(var i=0; i<this._opponents.length; i++) {
            if (this._opponents[i] === opponent) {
                alreadyPresent = true;
                break;
            }
        }
        if (!alreadyPresent) {
            this._opponents.push(opponent);
        }
    },
    resolveFights: function() {
        while(this._opponents.length > 0) {
            this._resolveFight(this._opponents.pop());
        }
    },
    _limitPerf: function(perf) {
        if (perf > 20) {
            return 20;
        } else if (perf < 0) {
            return 0;
        } else {
            return perf;
        }
    },
    _resolveFight: function(target) {
        var thisPerf = this._limitPerf(this.attackPerformance());
        this.sendSelfMessage("You got %s attack on the %s.",[Game.Rules.aadj(thisPerf), target.getName()]);
        this.sendMessage(target,"The %s got %s attack on you.",[this.getName(), Game.Rules.aadj(thisPerf)]);
        var thatPerf = this._limitPerf(target.attackPerformance());
        this.sendMessage(target,"You got %s attack on the %s.",[Game.Rules.aadj(thatPerf), this.getName()]);
        this.sendSelfMessage("The %s got %s attack on you.",[target.getName(), Game.Rules.aadj(thatPerf)]);
        if (thisPerf === thatPerf) {
            this.sendSelfMessage("You draw with the %s.",[target.getName()]);
            this.sendMessage(target,"You draw with the %s.",[this.getName()]);
            return false; // no one lost 
        }
        var winner = null;
        var loser = null;
        var difference = 0;
        if (thisPerf > thatPerf) {
            winner = this;
            loser = target;
            difference = thisPerf - thatPerf;
        } else {
            winner = target;
            loser = this;
            difference = thatPerf - thisPerf;
        }
        if (difference > 10) { difference = 10; }
        winner.setModifier(0);
        if (!winner.isDefensive()) {
            var died = loser.takeMorale(winner,difference);
            if (!died) {
                var damage = winner.getDamageScore();
                if (difference < 4) {
                    loser.sendMessage(winner,"You won only a minor victory over the %s.",[loser.getName()]);
                    winner.sendMessage(loser,"You suffered only a minor defeat from the %s.",[winner.getName()]);
                    loser.setModifier(-1);
                } else if (difference < 7) {
                    loser.sendMessage(winner,"You won a clear victory over the %s.",[loser.getName()]);
                    winner.sendMessage(loser,"You suffered a clear defeat from the %s.",[winner.getName()]);
                    loser.setModifier(-2);
                    died = loser.takeDamage(winner,Game.Rules.woundFromGap(difference,damage),0);
                } else if (difference < 10) {
                    loser.sendMessage(winner,"You won a major victory over the %s.",[loser.getName()]);
                    winner.sendMessage(loser,"You suffered a major defeat from the %s.",[winner.getName()]);
                    loser.setModifier(-3);
                    died = loser.takeDamage(winner,Game.Rules.woundFromGap(difference,damage),1);
                } else {
                    loser.sendMessage(winner,"You won a devastating victory over the %s.",[loser.getName()]);
                    winner.sendMessage(loser,"You suffered a devastating defeat from the %s.",[winner.getName()]);
                    loser.setModifier(-4);
                    died = loser.takeDamage(winner,Game.Rules.woundFromGap(difference,damage),2);
                }
            }
            if (died) {
                winner.raiseEvent("onKill",loser.getXP());
            }
            return;
        } else {
            this.sendMessage(winner,"You were fighting defensive and so didn't touch %s.",[loser.getName()]);
            this.sendMessage(loser,"The %s was fighting defensively and so didn't touch you.",[winner.getName()]);
            return;
        }
    }
};

Game.EntityMixins.TaskActor = {
    name: 'TaskActor',
    interfaceName: 'Actor',
    init: function(template) {
        this._tasks = template['tasks'] || ['wander'];
    },
    getTasks: function(tasks) {
        return this._tasks;
    },
    setTasks: function(tasks) {
        this._tasks = tasks;
    },
    act: function() {
        for (var i=0; i<this._tasks.length; i++) {
            if (this[this._tasks[i]]()) {
                return;
            }
        }
    },
    hunt:  function() {
        if (!this.hasMixin('Sight') || !this.canSee(this.getMap().getPlayer())) {
            return false;
        }
        
        var player = this.getMap().getPlayer();
        var playerx = player.getX();
        var playery = player.getY();
        
        if (Math.abs(playerx-this.getX()) <= 1 && Math.abs(playery-this.getY()) <= 1) {
            this.attack(player);
            return true;
        }
        
        var source = this;
        var z = source.getZ();
        var path = new ROT.Path.AStar(player.getX(), player.getY(), function(x,y) {
            var entity = source.getMap().getEntityAt(x,y,z);
            if (entity && entity !== source && entity !== player) {
                return false;
            }
            return source.getMap().getTile(x,y,z).isWalkable();
        },{topology:8});
        var count = 0;
        path.compute(source.getX(), source.getY(), function(x,y) {
            if (count === 1) {
                source.tryMove(x,y,z);
            }
            count++;
        });
        return (count > 0);
    },
    wander: function() {
        var moveOffset = (Math.round(Math.random()) === 1) ? 1 : -1;
        if (Math.round(Math.random()) === 1) {
            this.tryMove(this.getX() + moveOffset, this.getY(), this.getZ());
        } else {
            this.tryMove(this.getX(), this.getY() + moveOffset, this.getZ());
        }
        return true;
    }
};

Game.EntityMixins.Moody = {
    name: 'Moody',
    listeners: {
        onAttacked: function(attacker) {
            if (this.hasMixin('TaskActor')) {
                var tasks = this.getTasks();
                var hasHunt = false;
                for(var i=0; i<tasks.length; i++) {
                    if (tasks[i] === 'hunt') {
                        hasHunt = true;
                        break;
                    }
                }
                if (!hasHunt) {
                    tasks.unshift('hunt');
                    this.setTasks(tasks);
                }
            }
        }
    }
};

Game.EntityMixins.User = {
    name: 'User',
    listeners: {
        onUse: function(item) {
            if (item.isHealingMorale()) {
                this.raiseEvent("onHealMorale",item);
            }
            if (item.getHeals() !== "none") {
                this.raiseEvent("onHeal",item);
            }
        }
    }
};

Game.EntityMixins.Equipper = {
    name: 'Equipper',
    init: function() {
        this._swordHand = Game.FistItem;
        this._shieldHand = Game.FistItem;
        this._armour = Game.NullItem;
    },
    wieldSword: function(item) {
        this._swordHand = item;
    },
    unwieldSword: function() {
        this._swordHand = Game.FistItem;
    },
    wieldShield: function(item) {
        this._shieldHand = item;
    },
    unwieldShield: function() {
        this._shieldHand = Game.FistItem;
    },
    wear: function(item) {
        this._armour = item;
    },
    takeOff: function() {
        this._armour = Game.NullItem;
    },
    getSwordHand: function() {
        return this._swordHand;
    },
    getShieldHand: function() {
        return this._shieldHand;
    },
    getArmour: function() {
        return { leftLeg:   this._armour.getArmour().leftLeg   + this._shieldHand.getArmour().leftLeg   ,
                 rightLeg:  this._armour.getArmour().rightLeg  + this._shieldHand.getArmour().rightLeg  ,
                 abdomen:   this._armour.getArmour().abdomen   + this._shieldHand.getArmour().abdomen   ,
                 chest:     this._armour.getArmour().chest     + this._shieldHand.getArmour().chest     ,
                 swordArm:  this._armour.getArmour().swordArm  + this._shieldHand.getArmour().swordArm  ,
                 shieldArm: this._armour.getArmour().shieldArm + this._shieldHand.getArmour().shieldArm ,
                 head:      this._armour.getArmour().head      + this._shieldHand.getArmour().head      ,
                 encumbrance: this._armour.getEncumbrance()                                             ,
                 getEncumbrance: function() {
                   return this.encumbrance;
               }
        };
    },
    unequip: function(item) {
        if (this._swordHand === item) {
            this.unwieldSword();
        }
        if (this._shieldHand === item) {
            this.unwieldShield();
        }
        if (this._armour === item) {
            this.takeOff();
        }
    }
    
};